<?php

/**
 * etherFAX_Client implementation.
 * Copyright (c) 2009-2012, etherFAX, LLC All rights reserved.
 */

/**
 * FaxResult values.
 */
class FaxResult
{
	const Success = 0;
	const Error = 1;
	const InProgress = 2;
	const LineBusy = 3;
	const LineDead = 4;
	const LineFailure = 5;
	const NoDialTone = 6;
	const NoAnswer = 7;
	const InvalidOrMissingNumber = 8;
	const InvalidOrMissingFile = 9;
	const InvalidChannel = 10;
	const UnexpectedDisconnect = 11;
	const NoChannelsAvailable = 12;
	const ChannelUnavailable = 13;
	const NothingToCancel = 14;
	const DeviceTimeout = 15;
	const DeviceBusy = 16;
	const NotFaxMachine = 17;
	const IncompatibleFaxMachine = 18;
	const FileError = 19;
	const FileNotFound = 20;
	const FileUnsupported = 21;
	const CallCollision = 22;
	const Cancelled = 23;
	const CallBlocked = 24;
	const DestinationBlackListed = 25;

	const Unauthorized = 100;
	const InvalidParameter = 101;
	const NotImplemented = 102;
	const ItemNotFound = 103;
}

/**
 * FaxState values.
 */
Class FaxState
{
	const Idle = 0;
	const Initializing = 1;
	const Dialing = 2;
	const Answering = 3;
	const Negotiating = 4;
	const Sending = 5;
	const Receiving = 6;
	const Cancelling = 7;
	const Disconnecting = 8;
}

/**
 * FaxAccount class.
 */
class FaxAccount
{
	public $Account;
	public $Name;
	public $Status;
	public $Ports;
}

/**
 * FaxStatus class.
 */
class FaxStatus
{
	public $Result;
	public $State;
	public $JobId;
	public $PagesDelivered;
	public $ConnectTime;
	public $ConnectSpeed;
	public $Tag;
	public $CompletedOn;

	/**
	 * Default constructor.
	 */
	public function __construct()
	{
		$this->Result = 0;
		$this->State = FaxState::Idle;
		$this->PagesDelivered = 0;
		$this->ConnectTime = 0;
		$this->ConnectSpeed = 0;
	}
}

/**
 * FaxReceive class.
 */
class FaxReceive
{
	public $Result;
	public $JobId;
	public $CalledNumber;
	public $CallingNumber;
	public $RemoteId;
	public $PagesReceived;
	public $ConnectTime;
	public $ConnectSpeed;
	public $ReceivedOn;
	public $FaxImage;

	/**
	 * Default constructor.
	 */
	public function __construct(array $a = NULL, $image = FALSE)
	{
		// initialize if array
		if (!is_null($a))
		{
			$this->Result = intval($a['FaxResult']);
			$this->JobId = $a['JobId'];
			$this->CalledNumber = $a['CalledNumber'];
			$this->CallingNumber = $a['CallingNumber'];
			$this->RemoteId = $a['RemoteId'];
			$this->PagesReceived = intval($a['PagesReceived']);
			$this->ConnectTime = intval($a['ConnectTime']);
			$this->ConnectSpeed = intval($a['ConnectSpeed']);
			$this->ReceivedOn = $a['ReceivedOn'];

			if ($image == TRUE)
				$this->FaxImage = base64_decode($a['FaxImage']);
		}
	}
}

/**
 * EtherFax_Client class.
 */
class EtherFax_Client
{
	// class constants
	const EFAX_API_URL = 'https://na.connect.etherfax.net/web/api';
	const DEFAULT_TIMEOUT = 30;
	const HTTP_OK = 200;

	// private data
	protected $auth;
	protected $httpCode;
	protected $timeout;

	/**
	 * Constructor with optional credentials.
	 */
	public function __construct($user = NULL, $password = NULL)
	{
		// set credentials, default timeout
		$this->setCredentials($user, $password);
		$this->timeout = EtherFax_Client::DEFAULT_TIMEOUT;
	}

	/**
	 * Sets the http request timeout.
	 */
	public function setTimeout($timeout)
	{
		$this->timeout = $timeout;
	}

	/**
	 * Returns the $httpCode from the last operation.
	 */
	public function getHttpCode()
	{
		return $this->httpCode;
	}

	/**
	 * Sets the credentials to use when connecting to the etherFAX web service.
	 */
	public function setCredentials($user, $password)
	{
		// set credentials
		if (is_null($user))
			$this->auth = NULL;
		else
			$this->auth = $user . ':' . $password;
	}

	/**
	 * Hanldes the http request "get" method and returns the response. The $error
	 * variable is also set and accessible using getLastError().
	 */
	private function curl_get($url, array $get = NULL)
	{
		// create full uri request
		$uri = EtherFax_Client::EFAX_API_URL . $url;
		if ($get != NULL)
			$uri .= '?' . http_build_query($get);

		// set default options
		$defaults = array(CURLOPT_URL => $uri,
			CURLOPT_HEADER => 0,
			CURLOPT_USERPWD => $this->auth,
			CURLOPT_RETURNTRANSFER => TRUE,
			CURLOPT_TIMEOUT => $this->timeout);

		// execute web request
		$ch = curl_init();
		curl_setopt_array($ch, $defaults);
		$result = curl_exec($ch);
		$this->httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);

		// close and return result
		curl_close($ch);
		return $result;
	}

	/**
	 * Handles the http request "post" method and returns the response. The vars contains
	 * an array of items that will be url encoded.
	 */
	private function curl_post($url, array $post = NULL)
	{
		// create full uri
		$uri = EtherFax_Client::EFAX_API_URL . $url;

		// set default options
		$defaults = array(CURLOPT_URL => $uri,
			CURLOPT_POST => 1,
			CURLOPT_HEADER => 0,
			CURLOPT_USERPWD => $this->auth,
			CURLOPT_FRESH_CONNECT => 1,
			CURLOPT_RETURNTRANSFER => 1,
			CURLOPT_FORBID_REUSE => 1,
			CURLOPT_TIMEOUT => $this->timeout,
			CURLOPT_POSTFIELDS => http_build_query($post));

		// execute web request
		$ch = curl_init();
		curl_setopt_array($ch, $defaults);
		$result = curl_exec($ch);
		$this->httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);

		// close and return result
		curl_close($ch);
		return $result;
	}

	/**
	 * Gets account/status information.
	 */
	public function getAccountStatus()
	{
		$acc = new FaxAccount();

		// get account status
		$response = $this->curl_get('/account?a=status');
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			// process response
			$xml = simplexml_load_string($response);
			$array = json_decode(json_encode($xml), TRUE);

			// get fax account info
			$a = $array['AccountStatus'];
			$acc->Account = $a['Account'];
			$acc->Name = $a['Name'];
			$acc->Status = $a['Status'];
			$acc->Ports = intval($a['Ports']);
		}

		return $acc;
	}

	/**
	 * Sends a fax to the specified number with optional params.
	 */
	public function sendFax($number, $file, $pages, $localId = NULL, $callerId = NULL, $tag = NULL, $tz = NULL)
	{
		// create fax status
		$status = new FaxStatus();

		// get file contents
		$data = file_get_contents($file);
		if (strlen($data) == 0)
		{
			// indicate file error
			$status->Result = FaxResult::InvalidOrMissingFile;
			return $status;
		}

		// set default timezone offset
		if (is_null($tz))
			$tz = 0;

		// set default page count
		if (is_null($pages))
			$pages = 0;

		// create post array/items
		$post = array('DialNumber' => $number,
			'FaxImage' => base64_encode($data),
			'TotalPages' => $pages,
			'TimeZoneOffset' => $tz);

		// add optional items
		if (!is_null($localId))
			$post['LocalId'] = $localId;
		if (!is_null($callerId))
			$post['CallerId'] = $callerId;
		if (!is_null($tag))
			$post['Tag'] = $tag;

		// set error default
		$status->Result = FaxResult::Error;

		// send fax
		$response = $this->curl_post('/outbox', $post);
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			// process response
			$xml = simplexml_load_string($response);
			$array = json_decode(json_encode($xml), TRUE);

			// get fax result and job id
			$a = $array['FaxStatus'];
			$status->Result = intval($a['FaxResult']);
			$status->JobId = $a['JobId'];
		}

		// return fax status
		return $status;
	}

	/**
	 * Returns the status of the specified fax id.
	 */
	public function getFaxStatus($id)
	{
		// create fax status
		$status = new FaxStatus();

		// get fax status for specified id
		$post = array('ids' => $id);
		$response = $this->curl_post('/status', $post);
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			// process response
			$xml = simplexml_load_string($response);
			$array = json_decode(json_encode($xml), TRUE);

			// get fax status
			$a = $array['FaxStatus'];
			$status->Result = intval($a['FaxResult']);
			$status->State = intval($a['State']);
			$status->JobId = $a['JobId'];
			$status->ConnectTime = intval($a['ConnectTime']);
			$status->ConnectSpeed = intval($a['ConnectSpeed']);
			$status->PagesDelivered = intval($a['PagesDelivered']);
			$status->RemoteId = $a['RemoteId'];
//			$status->Tag = $a['Tag'];
		}

		// return status
		return $status;
	}

	/**
	 * Gets the number of pending faxes waiting for delivery.
	 */
	public function getPendingFaxCount()
	{
		// get pending fax count
		$response = $this->curl_get('/outbox?a=pending');
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			$xml = simplexml_load_string($response);
			return intval($xml->PendingFaxes);
		}
		return 0;
	}

	/**
	 * Gets the number of unread faxes.
	 */
	public function getUnreadFaxCount()
	{
		// get unread fax count
		$response = $this->curl_get('/inbox?a=unread');
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			$xml = simplexml_load_string($response);
			return intval($xml->UnreadFaxes);
		}
		return 0;
	}

	/**
	 * Gets a list of unread faxes.
	 */
	public function getUnreadFaxList()
	{
		// get unread fax list
		$response = $this->curl_get('/inbox?a=list');
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			$xml = simplexml_load_string($response);
			$array = json_decode(json_encode($xml), TRUE);

			// build receive fax list
			$faxes = array();
			foreach($array as $a)
			{
				$faxes[] = new FaxReceive($a);
			}

			// return fax list
			return $faxes;
		}

		return NULL;
	}

	/**
	 * Retrieves (downloads) the specified fax id including the image.
	 */
	public function getFax($id)
	{
		// retrieve the specified fax
		$response = $this->curl_get('/inbox?a=get&id=' . $id);
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
		{
			$xml = simplexml_load_string($response);
			$a = json_decode(json_encode($xml), TRUE);

			// return new fax receive
			return new FaxReceive($a['FaxReceive'], TRUE);
		}

		return NULL;
	}

	/**
	 * Sets the specified fax id as "received". Once called, this fax will no longer
	 * appear in the unread fax items.
	 */
	public function setFaxReceived($id)
	{
		// set fax as received
		$response = $this->curl_get('/inbox?a=received&id=' . $id);
		if ($response && $this->httpCode == EtherFax_Client::HTTP_OK)
			return true;

		return false;
	}
}

?>
