﻿#region Copyright
//
// EtherFaxRestClient Implementation
// Copyright (c) 2010-2015, etherFAX, LLC  All rights reserved.
//
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Net;
using System.Text;
using System.Web;
using System.Web.Script.Serialization;

namespace EtherFax.Web.Rest
{
	/// <summary>
	/// FaxResult values.
	/// </summary>
	public enum FaxResult
	{
		Success = 0,
		Error,
		InProgress,
		LineBusy,
		LineDead,
		LineFailure,
		NoDialTone,
		NoAnswer,
		InvalidOrMissingNumber,
		InvalidOrMissingFile,
		InvalidChannel,
		UnexpectedDisconnect,
		NoChannelsAvailable,
		ChannelUnavailable,
		NothingToCancel,
		DeviceTimeout,
		DeviceBusy,
		NotFaxMachine,
		IncompatibleFaxMachine,
		FileError,
		FileNotFound,
		FileUnsupported,
		CallCollision,
		Cancelled,
		CallBlocked,
		DestinationBlackListed,

		Unauthorized = 100,
		InvalidParameter,
		NotImplemented,
		ItemNotFound,
	}

	/// <summary>
	/// FaxState values.
	/// </summary>
	public enum FaxState
	{
		Idle = 0,
		Initializing,
		Dialing,
		Answering,
		Negotiating,
		Sending,
		Receiving,
		Cancelling,
		Disconnecting,
		Conversion
	}

	/// <summary>
	/// FaxStatus class.
	/// </summary>
	public class FaxStatus
	{
		public FaxResult FaxResult { get; set; }
		public FaxState State { get; set; }
		public Guid JobId { get; set; }
		public int PagesDelivered { get; set; }
		public int ConnectTime { get; set; }
		public int ConnectSpeed { get; set; }
		public string RemoteId { get; set; }
		public string Tag { get; set; }
		public DateTime? CompletedOn { get; set; }
	}

	/// <summary>
	/// FaxAccount class.
	/// </summary>
	public class FaxAccount
	{
		public string Account { get; set; }
		public int AccountId { get; set; }
		public int Country { get; set; }
		public string Name { get; set; }
		public bool Enabled { get; set; }
		public int Ports { get; set; }
		public int Numbers { get; set; }
		public string[] Features { get; set; }
	}

	/// <summary>
	/// FaxEventStatus class.
	/// </summary>
	public class FaxEventStatus
	{
		public bool FaxReceived { get; set; }
		public bool FaxCompleted { get; set; }
		public bool FaxStatus { get; set; }
	}

	/// <summary>
	/// FaxReceive class.
	/// </summary>
	public class FaxReceive
	{
		public FaxResult FaxResult { get; set; }
		public Guid JobId { get; set; }
		public string CalledNumber { get; set; }
		public string CallingNumber { get; set; }
		public string RemoteId { get; set; }
		public int PagesReceived { get; set; }
		public int ConnectTime { get; set; }
		public int ConnectSpeed { get; set; }
		public DateTime ReceivedOn { get; set; }
		public string FaxImage { get; set; }
	}

	/// <summary>
	/// PendingFaxes
	/// </summary>
	public class PendingFaxResult
	{
		public int PendingFaxes { get; set; }
		public int AvailableChannels { get; set; }
	}

	/// <summary>
	/// UnreadFaxResult
	/// </summary>
	public class UnreadFaxResult
	{
		public int UnreadFaxes { get; set; }
	}

	/// <summary>
	/// EtherFaxRestClient class.
	/// </summary>
	class EtherFaxRestClient
	{
		// string constants
		private const string REST_API_URL = "https://na.connect.etherfax.net/rest/{0:.0}/api";

		// private members
		private string _auth = null;
		private string _response = null;
		private int _timeout = 0;
		private decimal _ver;
		private string _base;
		private HttpStatusCode _statusCode = HttpStatusCode.ServiceUnavailable;
		private JavaScriptSerializer _json = new JavaScriptSerializer();

		/// <summary>
		/// Returns the http status code.
		/// </summary>
		public HttpStatusCode StatusCode { get { return _statusCode; } }

		/// <summary>
		/// Returns true is http result is 200/OK.
		/// </summary>
		public bool IsOK { get { return _statusCode == HttpStatusCode.OK; } }

		/// <summary>
		/// Gets the current service base address.
		/// </summary>
		public string ServiceAddress { get { return _base; } }

		/// <summary>
		/// Gets/sets the web client timeout.
		/// </summary>
		public int Timeout
		{
			get { return _timeout; }
			set { _timeout = value; }
		}

		/// <summary>
		/// Gets/sets the rest api version.
		/// </summary>
		public string Version
		{
			get { return _ver.ToString(".0"); }
			set
			{
				_ver = decimal.Parse(value);
				_base = string.Format(REST_API_URL, _ver);
			}
		}

		/// <summary>
		/// Gets the response string.
		/// </summary>
		public string Response { get { return _response; } }

		/// <summary>
		/// Default constructor.
		/// </summary>
		public EtherFaxRestClient()
		{
			// set default api version
			Version = "2";
		}

		/// <summary>
		/// Constructor with user/password.
		/// </summary>
		public EtherFaxRestClient(string user, string password) : this ()
		{
			SetCredentials(user, password);
		}

		/// <summary>
		/// Sets the credentials for the connection.
		/// </summary>
		public void SetCredentials(string user, string password)
		{
			// valid user?
			if (String.IsNullOrEmpty(user))
			{
				_auth = null;
				return;
			}

			// set empty password if null
			if (password == null)
				password = "";

			// create auth string
			_auth = Convert.ToBase64String(Encoding.ASCII.GetBytes(user + ":" + password));
		}

		/// <summary>
		/// Gets the fully qualified connection/method url.
		/// </summary>
		public string GetUrl(string url)
		{
			// create full url
			return url[0] == '/' ? _base + url : _base + "/" + url;
		}

		/// <summary>
		/// Creates request using url, method, and optional values.
		/// </summary>
		public string Request(string url, string method, object vars)
		{
			// create fully qualified url
			url = GetUrl(url);
			_response = null;

			// create web request
			HttpWebRequest req = (HttpWebRequest)HttpWebRequest.Create(url);
			if (_timeout > 0)
				req.Timeout = _timeout * 1000;
			req.Headers.Add("Authorization", "Basic " + _auth);
			req.Method = method.ToUpper();

			// post method?
			if (req.Method == "POST")
			{
				// post data?
				if (vars != null)
				{
					byte[] data = null;

					// hash table?
					if (vars.GetType() == typeof(Hashtable))
					{
						// encode variables if present
						StringBuilder sb = new StringBuilder();
						foreach (DictionaryEntry d in (Hashtable)vars)
							sb.Append(d.Key.ToString() + "=" +
							   HttpUtility.UrlEncode(d.Value.ToString()) + "&");

						// set request data
						data = Encoding.UTF8.GetBytes(sb.ToString());
						req.ContentType = "application/x-www-form-urlencoded";
						req.ContentLength = data.Length;
					}

					// json string?
					else if (vars.GetType() == typeof(string))
					{
						data = Encoding.UTF8.GetBytes((string)vars);
						req.ContentType = "application/json";
						req.ContentLength = data.Length;
					}

					// disable 100/continue
					ServicePointManager.Expect100Continue = false;

					// write to stream
					using (Stream rs = req.GetRequestStream())
					{
						rs.Write(data, 0, data.Length);
						rs.Close();
					}
				}
			}

			try
			{
				// get response
				HttpWebResponse resp = (HttpWebResponse)req.GetResponse();
				_statusCode = resp.StatusCode;
				if (resp.StatusCode == HttpStatusCode.OK)
				{
					using (Stream rs = resp.GetResponseStream())
					{
						using (StreamReader reader = new StreamReader(rs, Encoding.UTF8))
						{
							_response = reader.ReadToEnd();
							reader.Close();
						}

						rs.Close();
					}
				}

				// close response
				resp.Close();
			}
			catch (WebException e)
			{
				HttpWebResponse r = (HttpWebResponse)e.Response;
				_statusCode = r.StatusCode;
			}

			// return string response
			return _response;
		}

		/// <summary>
		/// Creates request using url and method only.
		/// </summary>
		public string Request(string url, string method)
		{
			return Request(url, method, null);
		}

		#region Account functions.

		/// <summary>
		/// Returns the current account status information.
		/// </summary>
		public FaxAccount GetFaxAccount()
		{
			FaxAccount status = null;

			try
			{
				// get account status
				string response = Request("accounts", "GET");

				// success?
				if (IsOK)
					status = _json.Deserialize<FaxAccount>(response);
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}

			// return account status result
			return status;
		}

		#endregion

		#region Outbox/send and status functions.

		/// <summary>
		/// Sends a fax to the specified number (full parameters).
		/// </summary>
		public FaxStatus SendFax(string number, string file, string localId, string callerId, string tag)
		{
			// create fax status
			FaxStatus status = new FaxStatus();

			int pages = 0;

			// load file
			byte[] data;

			try
			{
				// read all bytes from file
				data = File.ReadAllBytes(file);

				// document?
				if (IsDocument(data))
					pages = 0;

				// otherwise tiff
				else
				{
					// get number of pages in image
					pages = GetImagePageCount(data);
					if (pages == 0)
					{
						// indicate invalid file
						status.FaxResult = FaxResult.InvalidOrMissingFile;
						return status;
					}
				}
			}
			catch
			{
				// indicate file error
				status.FaxResult = FaxResult.FileError;
				return status;
			}

			// create value list
			Hashtable h = new Hashtable();

			// add dial number, fax image, and page count
			h.Add("DialNumber", number);
			h.Add("FaxImage", Convert.ToBase64String(data));
			h.Add("TotalPages", pages.ToString());

			// fax csid (optional)
			if (!string.IsNullOrEmpty(localId))
				h.Add("LocalId", localId);

			// caller id (optional)
			if (!string.IsNullOrEmpty(callerId))
				h.Add("CallerId", callerId);

			// application tag (optional)
			if (!string.IsNullOrEmpty(tag))
				h.Add("Tag", tag);

			// set utc offset (fax header date/time localization)
			TimeSpan span = TimeZone.CurrentTimeZone.GetUtcOffset(DateTime.Now);
			h.Add("TimeZoneOffset", span.Hours.ToString());

			status.FaxResult = FaxResult.Error;

			try
			{
				JavaScriptSerializer js = new JavaScriptSerializer();
				string json = js.Serialize(h);

				// send fax request
				string response = Request("outbox", "POST", json);

				// success?
				if (IsOK)
					status = _json.Deserialize<FaxStatus>(response);
			}
			catch (Exception e)
			{
				Console.WriteLine("send fax exception: " + e.Message);
			}

			// return etherfax status/result
			return status;
		}

		/// <summary>
		/// Returns the number of pending faxes for this account.
		/// </summary>
		public PendingFaxResult GetPendingFaxes()
		{
			PendingFaxResult result = null;

			try
			{
				// get account status
				string response = Request("outbox?a=pending", "GET");

				// success?
				if (IsOK)
				{
					result = _json.Deserialize<PendingFaxResult>(response);
				}
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}

			// return pending fax count
			return result;
		}

		/// <summary>
		/// Get fax status for single job id.
		/// </summary>
		public FaxStatus GetFaxStatus(Guid id)
		{
			FaxStatus status = null;

			try
			{
				// send fax request
				string response = Request("status?id=" + id.ToString(), "GET");

				// success?
				if (IsOK)
					status = _json.Deserialize<FaxStatus>(response);
			}
			catch (Exception e)
			{
				Console.WriteLine("status exception: " + e.Message);
			}

			return status;
		}

		/// <summary>
		/// Waits for any fax event and returns the fax event status.
		/// </summary>
		public FaxEventStatus WaitForFaxEvents(int wait)
		{
			return WaitForFaxEvents(wait, null);
		}

		/// <summary>
		/// Waits for fax events and returns the fax event status. On input, type may be
		/// receive|send|status|any. If type is null, all events are monitored.
		/// </summary>
		public FaxEventStatus WaitForFaxEvents(int wait, string type)
		{
			FaxEventStatus status = new FaxEventStatus();

			try
			{
				// send fax request
				string uri = string.Format("status?wait={0}&type={1}", wait, type != null ? type : "any");
				string response = Request(uri, "GET");

				// success?
				if (IsOK)
					status = _json.Deserialize<FaxEventStatus>(response);
			}
			catch
			{
			}

			return status;
		}

		/// <summary>
		/// Get fax status for multiple job ids at once.
		/// </summary>
		public List<FaxStatus> GetFaxStatus(List<Guid> ids)
		{
			try
			{
				// build job-id list
				StringBuilder sb = new StringBuilder();
				foreach (Guid id in ids)
				{
					if (sb.Length > 0)
						sb.Append(",");
					sb.Append(id.ToString());
				}

				// add post data
				Hashtable h = new Hashtable();
				h.Add("ids", sb.ToString());

				// send fax request
				string response = Request("status", "POST", h);

				// success?
				if (IsOK)
				{
					List<FaxStatus> list = _json.Deserialize<List<FaxStatus>>(response);
					return list;
				}
			}
			catch (Exception e)
			{
				Console.WriteLine("status exception: " + e.Message);
			}

			return null;
		}

		#endregion

		#region Inbox/receive fax functions.

		/// <summary>
		/// Returns unread fax count without wait parameter.
		/// </summary>
		public int GetUnreadFaxCount()
		{
			return GetUnreadFaxCount(0);
		}

		/// <summary>
		/// Returns the number of unread faxes for this account.
		/// </summary>
		public int GetUnreadFaxCount(int wait)
		{
			int unread = 0;

			try
			{
				// get unread fax count w/optional wait
				string response = Request(wait > 0 ?
					"inbox?a=unread&wait=" + wait : "inbox?a=unread", "GET");

				// success?
				if (IsOK)
				{
					UnreadFaxResult result = _json.Deserialize<UnreadFaxResult>(response);
					unread = result.UnreadFaxes;
				}
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}

			// return unread fax count
			return unread;
		}

		/// <summary>
		/// Returns a list of unread faxes (information only) up to maxCount. If maxCount
		/// is 0, all unread faxes are returned.
		/// </summary>
		public List<FaxReceive> GetUnreadFaxList(int maxCount)
		{
			List<FaxReceive> list = null;

			try
			{
				// get account status
				string response = Request("inbox?a=list&count=" + maxCount.ToString(), "GET");

				// success?
				if (IsOK)
					list = _json.Deserialize<List<FaxReceive>>(response);
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}

			// always return list (if empty)
			if (list == null)
				list = new List<FaxReceive>();

			return list;
		}

		/// <summary>
		/// Returns all unread faxes.
		/// </summary>
		public List<FaxReceive> GetUnreadFaxList()
		{
			return GetUnreadFaxList(0);
		}

		/// <summary>
		/// Gets (downloads) the specified fax id including the fax image.
		/// </summary>
		public FaxReceive GetFax(Guid id)
		{
			FaxReceive recv = null;

			try
			{
				// get account status
				string response = Request("inbox?a=get&id=" + id.ToString(), "GET");

				// success?
				if (IsOK)
					recv = _json.Deserialize<FaxReceive>(response);
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}

			return recv;
		}

		/// <summary>
		/// Gets (downloads) the next available fax waiting on the service.
		/// </summary>
		public FaxReceive GetNextUnreadFax(bool download = false)
		{
			FaxReceive recv = null;

			try
			{
				// get next unread fax w/ optional download image
				string response = Request("inbox?a=getnext&download=" + (string)(download ? "1" : "0"), "GET");

				// success?
				if (IsOK)
					recv = _json.Deserialize<FaxReceive>(response);
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}

			return recv;
		}

		/// <summary>
		/// Marks the specified fax as received.
		/// </summary>
		public void MarkFaxReceived(Guid id)
		{
			try
			{
				// mark fax as received
				Request("inbox?a=received&id=" + id.ToString(), "GET");
			}
			catch (Exception e)
			{
				Console.WriteLine("exception: " + e.Message);
			}
		}

		#endregion

		#region Class helper functions.

		/// <summary>
		/// Returns the number of pages in the specified image buffer.
		/// </summary>
		public int GetImagePageCount(byte[] image)
		{
			// upload image available?
			if (image == null)
				return 0;

			// set default page count
			int pages = 0;

			try
			{
				// stream image to determine page count
				using (MemoryStream ms = new MemoryStream(image))
				{
					using (Image img = Bitmap.FromStream(ms))
					{
						pages = img.GetFrameCount(FrameDimension.Page);
					}
				}
			}
			catch
			{
			}

			// return page count
			return pages;
		}

		/// <summary>
		/// Returns true if supported document format.
		/// </summary>
		private bool IsDocument(byte[] data)
		{
			try
			{
				// valid data?
				if (data != null && data.Length >= 4)
				{
					// pdf?
					if (data[0] == '%' && data[1] == 'P' && data[2] == 'D' && data[3] == 'F')
						return true;

					// rtf?
					if (data[0] == '{' && data[1] == '\\' && data[2] == 'r' && data[3] == 't' && data[4] == 'f')
						return true;
				}
			}
			catch
			{
			}

			return false;
		}

		#endregion
	}
}
