/**
 * EtherFaxClient.java
 * Copyright (c) 2009-2013, etherFAX LLC
 * 
 * This module provides a high-level function wrapper to send and receive faxes
 * using the etherFAX RESTful web service API.
 */

package etherfax.web;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.commons.codec.binary.Base64;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.message.BasicNameValuePair;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * EtherFaxClient
 */
public class EtherFaxClient {
	/**
	 * Constants
	 */
	private static final String SERVICE_URL = "https://na.connect.etherfax.net/web/api";
	public static final String VERSION = "1.0.1";
	
	/**
	 * Private data members
	 */
	private CloseableHttpClient _client;
	private String _auth;
	private String _response;
	private int _responseCode;
	
	/**
	 * Default constructor.
	 */
	public EtherFaxClient() {
	}

	/**
	 * Constructor with account, user, and password.
	 */
	public EtherFaxClient(String account, String user, String password) {
		setCredentials(account, user, password);
	}

	/**
	 * setCredentials
	 */
	public void setCredentials(String account, String user, String password) {
		// set authorization token
		String token = account + "/" + user + ":" + password;
		_auth = Base64.encodeBase64String(token.getBytes());
	}
	
	/**
	 * getResponse
	 */
	public String getResponse() {
		return _response;
	}

	/**
	 * getResponseCode
	 */
	public int getResponseCode() {
		return _responseCode;
	}

	/**
	 * getVersion
	 */
	public String getVersion() {
		return VERSION;
	}
	
	/**
	 * isOK
	 */
	public boolean isOK() {
		return _responseCode == 200;
	}

	/**
	 * create
	 */
	public boolean create() {
		try { 
			// create http client
			_client = HttpClients.createDefault();
			return true;
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		// indicate error
		return false;
	}
	
	/**
	 * close
	 */
	public void close() {
		try {
			// close client
			if (_client != null) {
				_client.close();
				_client = null;
			}
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * getUri
	 */
	private URI getUri(String uri)
	{
		String path = (uri.charAt(0) == '/') ? SERVICE_URL + uri : SERVICE_URL + "/" + uri;
		try {
			return new URI(path);
		}
		catch (URISyntaxException e) {
		}
		
		return null;
	}

	/**
	 * request
	 */
	public String request(String url, String method, HashMap<String, String> vars) throws Exception {
		// valid method?
		if (!method.equals("GET") && !method.equals("POST"))
			throw new Exception("Invalid method specified.");

		// reset response code
		_responseCode = 0;

		// get with arguments?
		if (method.equals("GET") && vars != null) {
			// create string builder
			StringBuilder sb = new StringBuilder(url);
			boolean sep = false;
			if (url.contains("?"))
				sep = true;
			else
				sb.append('?');

			// read all key/value pairs
			for (Map.Entry<String, String> entry : vars.entrySet()) {
				if (sep)
					sb.append('&');
				sb.append(entry.getKey() + "=" + URLEncoder.encode(entry.getValue(), "utf-8"));
				sep = true;
			}

			// set final url
			url = sb.toString();
		}

		// create response
		HttpResponse response;
		_response = null;
		
		try {
			// post?
			if (method.equals("POST")) {
				// create post request
				HttpPost request = new HttpPost(getUri(url));
				
				if (vars != null) {
					List<NameValuePair> params = new ArrayList<NameValuePair>();
					
					// read all key/value pairs
					for (Map.Entry<String, String> entry : vars.entrySet())
						params.add(new BasicNameValuePair(entry.getKey(), entry.getValue()));
					
					// add entity data to post
					UrlEncodedFormEntity entity = new UrlEncodedFormEntity(params);
					request.setEntity(entity);
				}
				
				// send request
				request.setHeader("Authorization", "Basic " + _auth);
				response = _client.execute(request);
			}
			
			// otherwise handle get
			else {
				// create and send request
				HttpGet request = new HttpGet(getUri(url));
				request.setHeader("Authorization", "Basic " + _auth);
				response = _client.execute(request);
			}
			
			// execute request
			_responseCode = response.getStatusLine().getStatusCode();
			
			// 200/401?
			if (_responseCode == HttpURLConnection.HTTP_OK || _responseCode == 401) {
				HttpEntity entity = response.getEntity();
				if (entity != null) {
					// get response stream
					InputStream is = response.getEntity().getContent();
					BufferedReader br = new BufferedReader(new InputStreamReader(is));
	
					// get response
					StringBuilder sb = new StringBuilder();
					String line;
					while ((line = br.readLine()) != null)
						sb.append(line);
					is.close();
	
					// set response string
					_response = sb.toString();
				}
			}
		} catch (IOException e) {
			System.out.println("request exception: " + e.getMessage());
		}

		// return response
		return _response;
	}

	/**
	 * request
	 */
	public String request(String url, String method) throws Exception {
		// request with no values
		return request(url, method, null);
	}
	
	/**
	 * getDocument
	 */
	private Document getDocument(String xml) {
		Document doc = null;

		try {
			// create xml document from input string
			DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			InputSource is = new InputSource();
			is.setCharacterStream(new StringReader(xml));
			doc = db.parse(is);
			doc.getDocumentElement().normalize();
		}
		catch (Exception e) {
			doc = null;
			e.printStackTrace();
		}

		// return document
		return doc;
	}
	
	/**
	 * getAccountStatus
	 * 
	 * Returns the account status for the authenticated user.
	 */
	public AccountStatus getAccountStatus() {
		AccountStatus status = null;
		
		try {
			// get account status
			request("/account?a=status", "GET");
			
			if (isOK()) {
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("AccountStatus");
				
				if (nodes.getLength() > 0) {
					// create account status
					status = new AccountStatus();
					status.fromXmlNode(nodes.item(0));
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		return status;
	}

	/**
	 * getPendingFaxes
	 * 
	 * Returns the number of pending faxes in the account's outbox.
	 */
	public int getPendingFaxes()
	{
		int pending = 0;

		try {
			// request pending fax count (outbound)
			request("outbox?a=pending", "GET");

			// success?
			if (isOK()) {
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("PendingFaxes");
				if (nodes.getLength() > 0) {
					pending = Integer.parseInt(nodes.item(0).getTextContent());
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// return pending fax count
		return pending;
	}
	
	/**
	 * getUnreadFaxCount
	 * 
	 * Returns the number of received faxes waiting for download.
	 */
	public int getUnreadFaxCount()
	{
		int unread = 0;

		try {
			// request unread fax count
			request("inbox?a=unread", "GET");

			// success?
			if (isOK()) {
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("UnreadFaxes");
				if (nodes.getLength() > 0) {
					unread = Integer.parseInt(nodes.item(0).getTextContent());
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// return pending fax count
		return unread;
	}
	
	/**
	 * getUnreadFaxList
	 * 
	 * Returns a list of received faxes waiting for download. This function does
	 * not return the actual fax content. The count argument may be used to limit the
	 * number of faxes returned in the list.
	 */
	public ArrayList<FaxReceive> getUnreadFaxList(int count) {
		ArrayList<FaxReceive> list = null;

		try {
			// request unread fax list
			request("inbox?a=list&count=" + count, "GET");

			// success?
			if (isOK()) {
				// create list
				list = new ArrayList<FaxReceive>();
				
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("FaxReceive");
				if (nodes.getLength() > 0) {
					for (int n = 0; n < nodes.getLength(); n++) {
						// create fax receive
						FaxReceive r = new FaxReceive();
						r.fromXmlNode(nodes.item(n), false);
						list.add(r);
					}
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// return pending fax count
		return list;
	}

	/**
	 * getUnreadFaxList
	 */
	public ArrayList<FaxReceive> getUnreadFaxList() {
		return getUnreadFaxList(100);
	}
	
	/**
	 * getFax
	 * 
	 * Returns the complete fax record for the specified job id including the
	 * fax content.
	 */
	public FaxReceive getFax(String id) {
		FaxReceive receive = null;

		try {
			// request unread fax list
			request("inbox?a=get&id=" + id, "GET");

			// success?
			if (isOK()) {
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("FaxReceive");
				if (nodes.getLength() > 0) {
					// create fax receive
					receive = new FaxReceive();
					receive.fromXmlNode(nodes.item(0), true);
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// return fax receive
		return receive;
	}

	/**
	 * setFaxReceived
	 * 
	 * Marks the specified fax id as received. This action removes the fax content
	 * permanently, but retains the call record deatil.
	 */
	public boolean setFaxReceived(String id) {
		try {
			// request pending fax count (outbound)
			request("inbox?a=received&id=" + id, "GET");

			// success?
			if (isOK()) {
				return true;
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
		return false;
	}

	/**
	 * sendFax
	 * 
	 * Send a document to the specified dial number with optional calling parameters.
	 */
	public FaxStatus sendFax(String number, String file, int pages, String localId, String callerId, String tag) {
		FaxStatus status = new FaxStatus();

		// load file
		byte[] data = null;

		try {
			// read all bytes from file
			File input = new File(file);
			FileInputStream is = new FileInputStream(file);
			data = new byte[(int)input.length()];
			is.read(data,  0,  data.length);
			is.close();
			
			// count pages?
			if (pages == 0) {
				// ...
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			
			// indicate file error
			status.Result = FaxResult.FileError;
			return status;
		}

		// create hash map
		LinkedHashMap<String, String> vars = new LinkedHashMap<String, String>();
		
		// add dial number, fax image, and page count
		vars.put("DialNumber", number);
		vars.put("FaxImage", Base64.encodeBase64String(data));
		vars.put("TotalPages", String.valueOf(pages));

		// fax csid (optional)
		if (localId != null)
			vars.put("LocalId", localId);

		// caller id (optional)
		if (callerId != null)
			vars.put("CallerId", callerId);

		// application tag (optional)
		if (tag != null)
			vars.put("Tag", tag);

		// set utc offset (fax header date/time localization)
		int offset = TimeZone.getDefault().getOffset(System.currentTimeMillis()) / (60 * 60 * 1000);
		vars.put("TimeZoneOffset",  String.valueOf(offset));

		status.Result = FaxResult.Error;

		try
		{
			// send fax request
			request("outbox", "POST", vars);

			// success?
			if (isOK())
			{
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("FaxStatus");
				if (nodes.getLength() > 0) {
					// create status
					status = new FaxStatus();
					status.fromXmlNode(nodes.item(0));
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// return etherfax status/result
		return status;
	}
	
	/**
	 * getFaxStatus
	 * 
	 * Returns the fax status for the specified job id.
	 */
	public FaxStatus getFaxStatus(String id) {
		FaxStatus status = null;
		
		try {
			// send status request
			request("status?id=" + id, "GET");

			// success?
			if (isOK()) {
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("FaxStatus");
				if (nodes.getLength() > 0) {
					// create status
					status = new FaxStatus();
					status.fromXmlNode(nodes.item(0));
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		return status;
	}
	
	/**
	 * getFaxStatus
	 * 
	 * Returns the fax status for the specified list of job ids.
	 */
	public ArrayList<FaxStatus> getFaxStatus(ArrayList<String> ids) {
		ArrayList<FaxStatus> list = new ArrayList<FaxStatus>();
		
		try {
			// build comma separated list
			StringBuilder sb = new StringBuilder();
			for (int n = 0; n < ids.size(); n++) {
				if (sb.length() > 0)
					sb.append(',');
				sb.append(ids.get(n));
			}
			
			// send status request
			request("status?id=" + sb.toString(), "GET");

			// success?
			if (isOK()) {
				// parse response
				Document d = getDocument(_response);
				NodeList nodes = d.getElementsByTagName("FaxStatus");
				if (nodes.getLength() > 0) {
					for (int n = 0; n < nodes.getLength(); n++) {
						// create status
						FaxStatus status = new FaxStatus();
						status.fromXmlNode(nodes.item(n));
						list.add(status);
					}
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		return list;
	}
}
